from __future__ import annotations

from pyJJAsim.compute import Matrix
from pyJJAsim.embedded_graph import EmbeddedGraph, EmbeddedTriangularGraph, EmbeddedHoneycombGraph, EmbeddedSquareGraph

import numpy as np
import scipy
import scipy.sparse
import scipy.sparse.linalg
import scipy.spatial

__all__ = ["Circuit", "SquareArray", "HoneycombArray", "TriangularArray", "SQUID"]

# TODO:
# work out if area and frustration definition
# give error if criterion all-or-nothing inductance is not met.
# give warning if timestep too small in low-inductance-case

# requirements
# - check/debug all physical quantities
# - dynamic compute broadcasting slicing
# - documentation
# - unit testing
# - publish


# nice to haves
# - lattices
# - periodic lattices
# - 3D
# - Multigrid
# - Nonlinear resistors
# - mixed inductor/non-inductor elements
# - GPU
#   * single/double precision
#   * phase build-up rounding correction


class NoCurrentConservationError(Exception):
    pass

class Circuit:
    """
    Construct a Josephson Circuit, also called a Josephson Junction Array (JJA).

    A JJA is an electric circuit that can include Josephson junctions, passive components,
    current sources and voltage sources. The network is required to be a planar embedding
    and single component, so junctions cannot intersect.

    Defined with a graph of nodes and edges where each edge contains a junction.
    A junction is the basic 2-terminal element which contains one of each component,
    see the included user manual for the precise definition. To omit a component;
    set it's value to zero.

    All physical quantities are normalized in pyjjasim, see the user manual for details.
    For example the critical current of each junction in Ampere is critical_current_factors * I0,
    where I0 is the normalizing scalar for all current values.

    Attributes
    ----------
    graph: EmbeddedGraph
        EmbeddedGraph instance with Nn nodes, Nj edges and Nf faces
    critical_current_factors=1.0:
        critical current factors of junctions, must be (Nj,) array or
        scalar (same value for all junctions)
    resistance_factors=1.0:
        resistance factors of junctions, must be (Nj,) array or
        scalar (same value for all junctions)
    capacitance_factors=0.0:
        capacitance factors of junctions, must be (Nj,) array or
        scalar  (same value for all junctions)
    inductance_factors=0.0:
        L_ij coupling between junction i and j. Either (Nj, Nj) sparse matrix,
        (Nj, Nj) array, (Nj,) array (entries of diagonal matrix) or
        scalar (constant diagonal)
    matrix_format="csc" :
        Format of sparse problem matrices, "csr" or "csc"

    Methods
    -------
    add_nodes_and_junctions(x, y, node1, node2, ...)
        Add new nodes and junctions to circuit. The Nn_new nodes are assigned
        ids range(Nn, Nn+Nn_new). Junction ids are asigned the same way.
    remove_nodes(nodes_to_remove)
        Remove nodes with a list/array of node ids of with a mask of shape (Nn,)
    remove_junctions(junctions_to_remove)
        Remove junctions with a list/array of junction ids of with a mask of shape (Nj,)
    get_node_coordinates()
        Returns x, y representing the coordinates of the nodes in the circuit.
    node_count()
        Returns the number of nodes in the circuit (abbreviated Nn)
    get_junction_count()
        Returns the number of junctions in the circuit (abbreviated Nj)
    get_face_count()
        Returns the number of faces in the circuit (abbreviated Nf)
    get_critical_current_factors()
        Returns the critical current factors assigned to each junction in the circuit
    get_resistance_factors()
        Returns the resistance factors assigned to each junction in the circuit
    get_capacitance_factors()
        Returns the capacitance factors assigned to each junction in the circuit
    get_inductance_factors()
        Returns the inductance factors, which is a matrix containing both
        self-inductances and mutual inductances between junctions.
    set_critical_current_factors(Ic)
        Modify the critical current factors of all junctions in the circuit
    set_resistance_factors(R)
        Modify the resistance factors of all junctions in the circuit
    set_capacitance_factors(C)
        Modify the capacitance factors of all junctions in the circuit
    set_inductance_factors(L)
        Modify the inductances factors of all junctions in the circuit
    get_faces(to_list=True)
        Returns a list of all faces, where a face is defined as an array
        containing ids of nodes encountered when traversing the boundary
        of a face counter-clockwise.
    locate_faces(x, y)
        Returns id of face at specified coordinate (where id is defined as
        the position in the list returned by get_faces)
    get_face_areas()
        Returns area of all faces in the circuit.
    get_face_centroids()
        Returns coordinates of the centroids of all faces in the circuit.
    approximate_inductance(junc_L=1, junc_M=0, max_dist=0)
        Construct inductance_factors matrix approximating geometric inductance
        in the array. junc_L on diangonals, junc_M * l1 * l2 * cos(alpha) / r
        on off-diagonals where alpha is angle between junctions, r is
        distance between centres of junctions and l1, l2 are lengths of junctions.
    get_cut_matrix(omit_last_node=False)
        Returns sparse cut matrix (shape (Nn, Nj), abbreviated M). Represents
        Kirchhoffs current law M @ I = 0
    get_cycle_matrix()
        Returns sparse cycle matrix (shape (Nf, Nj) abbreviated A). Represents
        Kirchhoffs voltage law A @ V = 0
    plot(...)
        Plot schematic representation of circuit. Can be used to visualise ids of
        nodes, junctions and faces.

    """

    def __init__(self, graph: EmbeddedGraph, critical_current_factors=1.0, resistance_factors=1.0,
                 capacitance_factors=0.0, inductance_factors=0.0, matrix_format="csc"):

        self.graph = graph
        self.graph.get_faces()
        n, _ = self.graph.get_face_nodes()
        self.graph.permute_faces(np.argsort(n[self.graph.faces_v_array.cum_counts]))
        self.graph.assert_planar_embedding()
        self.graph.assert_single_component()
        self.junction_unsort_perm = np.argsort(self.graph.edge_sorter)  # needed because EmbeddedGraph sorts edges.

        self.resistance_factors = None
        self.capacitance_factors = None
        self.critical_current_factors=None
        self.inductance_factors = None
        self.set_resistance_factors(resistance_factors)
        self.set_critical_current_factors(critical_current_factors)
        self.set_capacitance_factors(capacitance_factors)
        self.set_inductance_factors(inductance_factors)

        self.locator = None

        self.matrix_format = matrix_format
        self.cut_matrix = self.graph.cut_space()[:, self.junction_unsort_perm]
        self.cycle_matrix = self.graph.cycle_space(include_boundary_faces=False)[:, self.junction_unsort_perm]
        self.cut_reduced_square = None
        self.cut_matrix_reduced = None
        self.cut_square = None
        self.cut_matrix_reduced_transposed = None
        self.cut_matrix_transposed = None
        self.cycle_matrix_transposed = None
        self.cycle_square = None
        self._Mnorm = None
        self._Anorm = None

    def get_junction_nodes(self):
        return self.graph.node1[self.junction_unsort_perm], self.graph.node2[self.junction_unsort_perm]

    def get_juncion_coordinates(self):
        x, y = self.get_node_coordinates()
        n1, n2 = self.get_junction_nodes()
        return x[n1], y[n1], x[n2], y[n2]

    # noinspection PyArgumentList
    def copy(self):
        n1, n2 = self.get_junction_nodes()
        return Circuit(EmbeddedGraph(self.graph.x, self.graph.y, n1, n2),
                       critical_current_factors=self.get_critical_current_factors(),
                       resistance_factors=self.get_resistance_factors(),
                       capacitance_factors=self.get_capacitance_factors(),
                       inductance_factors=self.get_inductance_factors(),
                       matrix_format=self.matrix_format)

    # noinspection PyArgumentList
    def add_nodes_and_junctions(self, x, y, node1, node2,
                                critical_current_factors=1.0, resistance_factors=1.0,
                                capacitance_factors=1.0, inductance_factors=1.0):
        """ Add nodes to array.

        """
        """ Add junctions to array.
        in:  node1, node2                         int arrays in range(node_count)
             critical_current_factors=1.0:        float (Nj_new,) array  or  scalar
             resistance_factors=1.0:              positive float (Nj_new,) array  or  scalar
             capacitance_factors=1.0:             positive float (Nj_new,) array  or  scalar
        out: new_array                            Array
        """
        x = np.array(x, dtype=np.double).flatten()
        new_x = np.append(self.graph.x, x)
        new_y = np.append(self.graph.y, np.array(y, dtype=np.double).flatten())
        n1, n2 = self.get_junction_nodes()
        new_node1 = np.append(n1, np.array(node1, dtype=int).flatten())
        new_node2 = np.append(n2, np.array(node2, dtype=int).flatten())
        new_Ic = Matrix(self.critical_current_factors).stack(Matrix(critical_current_factors)).A
        new_R = Matrix(self.resistance_factors).stack(Matrix(resistance_factors)).A
        new_C = Matrix(self.capacitance_factors).stack(Matrix(capacitance_factors)).A
        new_L = Matrix(self.inductance_factors).stack(Matrix(inductance_factors)).A
        return Circuit(EmbeddedGraph(new_x, new_y, new_node1, new_node2),
                       critical_current_factors=new_Ic, resistance_factors=new_R,
                       capacitance_factors=new_C, inductance_factors=new_L,
                       matrix_format=self.matrix_format)

    # noinspection PyArgumentList
    def remove_nodes(self, nodes_to_remove):
        """ Remove nodes from array.
        in:  nodes_to_remove    int array in range(node_count)  or  (Nn,) mask
        out: new_array          Array
        """
        nodes_to_remove = np.array(nodes_to_remove).flatten()
        if not len(nodes_to_remove) == self.node_count():
            nodes_to_remove = np.array(nodes_to_remove, dtype=int)
        if not isinstance(nodes_to_remove.dtype, (bool, np.bool)):
            try:
                node_remove_mask = np.zeros(self.node_count(), dtype=bool)
                node_remove_mask[nodes_to_remove] = True
            except:
                raise ValueError("Invalid nodes_to_remove; must be None, mask, slice or index array")
        else:
            node_remove_mask = nodes_to_remove
        new_x = self.graph.x[~node_remove_mask]
        new_y = self.graph.y[~node_remove_mask]
        n1, n2 = self.get_junction_nodes()
        junc_remove_mask, new_node_id = self._junction_remove_mask(n1, n2, node_remove_mask)
        new_node1 = new_node_id[n1][~junc_remove_mask]
        new_node2 = new_node_id[n2][~junc_remove_mask]
        new_Ic = Matrix(self.critical_current_factors).select(~junc_remove_mask).A

        new_R = Matrix(self.resistance_factors).select(~junc_remove_mask).A
        new_C = Matrix(self.capacitance_factors).select(~junc_remove_mask).A
        new_L = Matrix(self.inductance_factors).select(~junc_remove_mask).A
        return Circuit(EmbeddedGraph(new_x, new_y, new_node1, new_node2),
                       critical_current_factors=new_Ic, resistance_factors=new_R,
                       capacitance_factors=new_C, inductance_factors=new_L,
                       matrix_format=self.matrix_format)

    # noinspection PyArgumentList
    def remove_junctions(self, junctions_to_remove):
        """ Remove junctions from array.
        in:  junctions_to_remove    int array in range(junction_count)  or  (Nj,) mask
        out: new_array              Array
        """
        junctions_to_remove = np.array(junctions_to_remove).flatten()
        if not len(junctions_to_remove) == self.junction_count():
            junctions_to_remove = np.array(junctions_to_remove, dtype=int)
        if not isinstance(junctions_to_remove.dtype, (bool, np.bool)):
            try:
                junction_mask = np.zeros(self.junction_count(), dtype=bool)
                junction_mask[junctions_to_remove] = True
            except:
                raise ValueError("Invalid junctions_to_remove; must be None, mask, slice or index array")
        else:
            junction_mask = junctions_to_remove
        n1, n2 = self.get_junction_nodes()
        new_node1, new_node2 = n1[~junction_mask], n2[~junction_mask]
        new_Ic = Matrix(self.critical_current_factors).select(~junction_mask).A
        new_R = Matrix(self.resistance_factors).select(~junction_mask).A
        new_C = Matrix(self.capacitance_factors).select(~junction_mask).A
        new_L = Matrix(self.inductance_factors).select(~junction_mask).A
        return Circuit(EmbeddedGraph(self.graph.x, self.graph.y, new_node1, new_node2),
                       critical_current_factors=new_Ic, resistance_factors=new_R,
                       capacitance_factors=new_C, inductance_factors=new_L,
                       matrix_format=self.matrix_format)

    def get_node_coordinates(self):
        """ Get node coordinates x and y
        out: x      (Nn,) float array
             y      (Nn,) float array
        """
        return self.graph.x, self.graph.y

    def node_count(self):
        return self._Nn()

    def get_critical_current_factors(self):
        return self.critical_current_factors

    def set_critical_current_factors(self, Ic):
        self.critical_current_factors = self._prepare_junction_quantity(Ic, self.junction_count(), x_name="Ic")
        return self

    def get_resistance_factors(self):
        return self.resistance_factors

    def set_resistance_factors(self, R):
        self.resistance_factors = self._prepare_junction_quantity(R, self.junction_count(), x_name="R")
        if np.any(self.resistance_factors <= 0.0):
            raise ValueError("All junctions must have a positive resistor")
        return self

    def get_capacitance_factors(self):
        return self.capacitance_factors

    def set_capacitance_factors(self, C):
        self.capacitance_factors = self._prepare_junction_quantity(C, self.junction_count(), x_name="C")
        if np.any(self.capacitance_factors < 0.0):
            raise ValueError("Capacitance cannot be negative.")
        return self

    def junction_count(self):
        return self._Nj()

    def face_count(self):
        return self._Nf()

    def get_faces(self, to_list=True):
        # Returns a list of faces.
        # if to_list==True returns in format:  [[n11, n12, n13], [n21], [n31, n32]]
        # if to_list==False returns in format: [n11, n12, n13, n21, n31, n32], [3, 1, 2]
        return self.graph.get_face_nodes(include_boundary_faces=False, to_list=to_list)

    def get_face_areas(self):
        # Returns unsigned area of each face in array.
        # out: areas            (face_count,) positive float array
        return self.graph.get_areas(include_boundary_faces=False)

    def get_face_centroids(self):
        # Returns centroid face_x, face_y of each face in array.
        # out: face_x, face_y   (face_count,) float arrRuehliay
        return self.graph.get_centroids(include_boundary_faces=False)

    def locate_faces(self, x, y):
        """ Get faces whose centroids are closest to queried (x,y) coordinate pairs.
        in:  x, y               float arrays of shape (N,)
        out: face_ids           int array in range(face_count) of shape (N,)
        """
        if self.locator is None:
            self.locator = scipy.spatial.KDTree(np.stack(self.get_face_centroids(), axis=-1))
        _, faces = self.locator.query(np.stack(np.broadcast_arrays(x, y), axis=-1), k=1)
        return faces

    def approximate_inductance(self, factor, junc_L=1, junc_M=0, max_dist=3):
        """
        Sets self.junction_inductance_factors (which is a Nj by Nj matrix) according to the formula:

         - self:     L = junc_L * l                                             in units of mu0 * a0
         - mutual:   L = junc_M * l_1 * l_2 * cos(alpha_12) / distance_12       in units of mu0 * a0
         -> self.junction_inductance_factors = factor * L                       -> in units of L0

        as a crude approximation of Neumann's formula for two wire segments.

        Here:
         - l, l1, l2    junction lengths                                        in units of a0
         - alpha_12     angle between junction 1 and junction 2
         - distance_12  distance between centres of junctions 1 and 2           in units of a0
         - max_dist     cutoff distance for included junction pairs             in units of a0

        junction_inductance_factors must be set in units of L0, so factor = mu0*a0/L0.
        """

        self.inductance_factors = None
        i, j = np.arange(self._Nj(), dtype=int), np.arange(self._Nj(), dtype=int)
        data = self._junction_lengths() * junc_L
        if junc_M > 0 and max_dist > 0:
            tree = scipy.spatial.KDTree(np.stack(self._junction_centers(), axis=-1))
            pairs = tree.query_pairs(max_dist, 2, output_type='ndarray')
            i, j = np.append(i, pairs[:, 0]), np.append(j, pairs[:, 1])
            i, j = np.append(i, pairs[:, 1]), np.append(j, pairs[:, 0])
            inner = self._junction_inner(*pairs.T)
            distance = self._junction_distance(*pairs.T)
            mutual = junc_M * inner / distance
            data = np.append(data, mutual)
            data = np.append(data, mutual)
        self.set_inductance_factors(factor * scipy.sparse.coo_matrix((data, (i, j)), shape=(self._Nj(), self._Nj())).tocsr())
        return self

    def get_inductance_factors(self):
        # return matrix whose entry (r, c) is the magnetic coupling between wire r and wire c.
        # out: (junction_count, junction_count) sparse symmetric float matrix
        return self.inductance_factors

    def set_inductance_factors(self, inductance_factors):
        # in: (junction_count, junction_count) sparse symmetric float matrix
        self.inductance_factors = inductance_factors
        L = Matrix(inductance_factors, assert_square=True, assert_symmetric=True)
        if not L.is_zero():
            eigv = scipy.sparse.linalg.eigsh(-L.matrix(self._Nj()).astype(np.double), 1, maxiter=1000, which="LA")[0][0]
            print(eigv)
            print(10 * np.finfo(float).eps)
            is_positive_definite = eigv < 100 * np.finfo(float).eps
            if not is_positive_definite:
                raise ValueError("Inductance matrix not positive definite")
        return self

    def __str__(self):
        return "x: \n" + str(self.graph.x) +"y: \n" + str(self.graph.y) + \
               "\nnode1: \n" + str(self.graph.node1) + "\nnode2: \n" + str(self.graph.node2)

    def get_cut_matrix(self):
        """
        returns the "cut-matrix" (abbreviated M in this code), which is a node_count by junction_count matrix.
        Its transpose is  the incidence matrix, node1 of a junction is -1 and node2 is -1 (abbreviated MT).
        Its rows span the cut-space (which is orthogonal to the cycle space). The cut-space has rank node_count - 1.

        Important: The row number does nót correspond with node_idx, because the nodes are permuted internally.

         out: (node_count, junction_count) sparse matrix
        """
        return self.cut_matrix

    def get_cycle_matrix(self):
        """
        returns the "cycle-matrix" (abbreviated A in this code), which is a face_count by junction_count matrix.
        It is +1 if traversing a face counter-clockwise passes through a junction in its direction, and -1
        otherwise. Its rows span the cycle-space (which is orthogonal to the cut space). The cut-space has a
        span equal to face_count = junction_count - node_count + 1.

        out: (face_count, junction_count) sparse matrix
        """
        return self.cycle_matrix

    def plot(self, show_node_ids=True, show_junction_ids=False, show_faces=True,
             show_face_ids=True, face_shrink_factor=0.9, figsize=None):
        """
        Visualize array. Can show nodes, junctions and faces; and their respective indices.
        For documentation see EmbeddedGraph.plot()
        """
        cr = self.graph.plot(show_faces=show_faces, figsize=figsize,
                             show_node_ids=show_node_ids, show_edge_ids=show_junction_ids,
                             show_face_ids=show_face_ids, face_shrink_factor=face_shrink_factor,
                             show_boundary_face=False)
        return cr

    # abbreviations and aliases
    def _Nn(self):   # alias for node_count
        return self.graph.node_count()

    def _Nnr(self):
        # reduced node count; returns node_count - 1
        return self._Nn() - 1

    def _Nj(self):   # alias for get_junction_count()
        return self.graph.edge_count()

    def _Nf(self):
        return self.graph.face_count(include_boundary_faces=False)

    def _Ic(self) -> Matrix:     # alias for get_critical_current_factors
        return Matrix(self.critical_current_factors, assert_square=True)

    def _R(self) -> Matrix:      # alias for get_resistance_factors
        return Matrix(self.resistance_factors, assert_square=True)

    def _C(self) -> Matrix:      # alias for get_resistance_factors
        return Matrix(self.capacitance_factors, assert_square=True)

    def _L(self) -> Matrix:
        return Matrix(self.inductance_factors, assert_square=True)

    def _Mr(self) -> Matrix:
        if self.cut_matrix_reduced is None:
            self.cut_matrix_reduced = self.cut_matrix[:-1, :]
        return Matrix(self.cut_matrix_reduced)

    def get_A_norm(self):
        # return ||A||_2 = sqrt(max(eig(A.T @ A))). (however computes sqrt(max(eig(A @ A.T))) which seems to be the same and is quicker)
        if self._Anorm is None:
            A = self.get_cycle_matrix()
            self._Anorm = np.sqrt(scipy.sparse.linalg.eigsh((A @ A.T).astype(np.double), 1, maxiter=1000, which="LA")[0][0])
        return self._Anorm

    def get_M_norm(self):
        # return ||M||_2 = sqrt(max(eig(M.T @ M))). (however computes sqrt(max(eig(M @ M.T))) which seems to be the same and is quicker)
        if self._Mnorm is None:
            M = self.get_cut_matrix()
            self._Mnorm = np.sqrt(scipy.sparse.linalg.eigsh((M @ M.T).astype(np.double), 1, maxiter=1000, which="LA")[0][0])
        return self._Mnorm

    def _area(self) -> Matrix:
        return Matrix(self.get_face_areas())

    def _A_solve(self, b):
        """
        Solves the equation: A @ x = b (where A = cycle_matrix).
        If b is integral (contain only integers), the output array x will also be integral.

        input:  b (..., Nf)
        output: x (..., Nj)

        Notes:
            - The equation is underdetermined, so the solution x is not unique.

        Use cases:
            - Used for changing phase zones (theta, z) -> (theta', z').
              Here theta' = theta + 2 * pi * Z where A @ Z = z' - z. Crucially, Z must
              be integral to ensure theta keeps obeying Kirchhoff's current rule.
            - Used for projecting theta onto cycle space; theta' = theta - g so that A @ theta'= 0.
              Then A @ g = 2 * pi * (z - areas * f)
        """
        return self.graph._cycle_space_solve_for_integral_x(b)[..., self.junction_unsort_perm]


    def _has_capacitance(self):
        # returns False if self.capacitance_factors is zero, True otherwise
        C = self._C()
        if C.is_scalar:
            if C.A == 0:
                return False
        return True

    def _has_inductance(self):
        # returns False if self.inductance_factors is zero, True otherwise
        L = self._L()
        if L.is_scalar:
            if L.A == 0:
                return False
        return True

    def _has_mixed_inductance(self):
        mask = self._get_mixed_inductance_mask()
        return np.any(mask) and not np.all(mask)

    def _get_mixed_inductance_mask(self):
        L = self._L().matrix(self._Nj())
        A = self.get_cycle_matrix()
        ALA = A @ L @ A.T
        return np.isclose(np.array(np.sum(np.abs(ALA), axis=1))[:, 0], 0)

    def _assign_cut_matrix(self):
        self.cut_square = None
        self.cut_matrix_reduced_transposed = None
        self.cut_matrix_transposed = None
        if self.cut_matrix_reduced is None or self.cut_matrix is None:
            cut_matrix = -self.graph.cut_space()
            self.cut_matrix = cut_matrix.asformat(self.matrix_format)
            self.cut_matrix_reduced = cut_matrix[:-1, :].asformat(self.matrix_format)
            return self.cut_matrix, self.cut_matrix_reduced

    @staticmethod
    def _prepare_junction_quantity(x, Nj, x_name="x"):
        x = np.array(x, dtype=np.double)
        if (x.ndim > 1) or not ((x.size == 1) or (x.size == Nj)):
            raise ValueError(x_name + " must be scalar or array of length Nj")
        return x

    def _junction_centers(self):
        x, y = self.get_node_coordinates()
        return 0.5 * (x[self.graph.node1] + x[self.graph.node2]),  0.5 * (y[self.graph.node1] + y[self.graph.node2])

    def _junction_lengths(self):
        x, y = self.get_node_coordinates()
        return np.sqrt((x[self.graph.node2] - x[self.graph.node1]) ** 2 + (y[self.graph.node2] - y[self.graph.node1]) ** 2)

    def _junction_inner(self, ids1, ids2):
        x, y = self.get_node_coordinates()
        x_n1_j1, y_n1_j1 = x[self.graph.node1[ids1]], y[self.graph.node1[ids1]]
        x_n2_j1, y_n2_j1 = x[self.graph.node2[ids1]], y[self.graph.node2[ids1]]
        x_n1_j2, y_n1_j2 = x[self.graph.node1[ids2]], y[self.graph.node1[ids2]]
        x_n2_j2, y_n2_j2 = x[self.graph.node2[ids2]], y[self.graph.node2[ids2]]
        return (x_n2_j1 - x_n1_j1) * (x_n2_j2 - x_n1_j2) + (y_n2_j1 - y_n1_j1) * (y_n2_j2 - y_n1_j2)

    def _junction_distance(self, ids1, ids2):
        x, y = self._junction_centers()
        return np.sqrt((x[ids2] - x[ids1]) ** 2 + (y[ids2] - y[ids1]) ** 2)

    @staticmethod
    def _junction_remove_mask(nodes1, nodes2, node_remove_mask):
        node_remove_mask = node_remove_mask.copy().astype(int)
        remove_nodes = np.flatnonzero(node_remove_mask)
        new_node_id = np.arange(node_remove_mask.size, dtype=int) - (np.cumsum(node_remove_mask) - node_remove_mask)
        junc_remove_mask = (np.isin(nodes1, remove_nodes) | np.isin(nodes2, remove_nodes))
        return junc_remove_mask, new_node_id

    @staticmethod
    def _lobpcg_matrix_norm(A, preconditioner=None, maxiter=1000, tol=1E-5):
        """
        Computes ||A||_2
        """
        x0 = np.random.rand(A.shape[0], 1)
        lobpcg_out = scipy.sparse.linalg.lobpcg(A, x0, B=None, M=preconditioner, maxiter=maxiter, tol=tol)
        return np.sqrt(lobpcg_out[0])


class SquareArray(Circuit):

    def __init__(self, count_x, count_y, x_scale=1.0, y_scale=1.0, matrix_format="csc"):
        super().__init__(EmbeddedSquareGraph(count_x, count_y, x_scale, y_scale),
                         matrix_format=matrix_format)

    def horizontal_junctions(self):
        x1, y1, x2, y2 = self.get_juncion_coordinates()
        return y1 == y2

    def vertical_junctions(self):
        return ~self.horizontal_junctions()


class HoneycombArray(Circuit):

    def __init__(self, count_x, count_y, x_scale=1.0, y_scale=1.0, matrix_format="csc"):
        super().__init__(EmbeddedHoneycombGraph(count_x, count_y, x_scale, y_scale),
                         matrix_format=matrix_format)

    def horizontal_junctions(self):
        x1, y1, x2, y2 = self.get_juncion_coordinates()
        return (y1 == y2).astype(int) * np.sign(x2 - x1)

    def vertical_junctions(self):
        return ~self.horizontal_junctions()


class TriangularArray(Circuit):

    def __init__(self, count_x, count_y, x_scale=1.0, y_scale=1.0, matrix_format="csc"):
        super().__init__(EmbeddedTriangularGraph(count_x, count_y, x_scale, y_scale),
                         matrix_format=matrix_format)

    def horizontal_junctions(self):
        x1, y1, x2, y2 = self.get_juncion_coordinates()
        return y1 == y2

    def vertical_junctions(self):
        return ~self.horizontal_junctions()


class SQUID(Circuit):

    """
    A SQUID is modeled as a square where the vertical junctions have Ic=1000
    and the horizontal Ic=1.
    """
    def __init__(self):
        x = [0, 1, 1, 0]
        y = [0, 0, 1, 1]
        node1 = [0, 1, 2, 0]
        node2 = [1, 2, 3, 3]
        graph = EmbeddedGraph(x, y, node1, node2)
        Ic = [1, 1000, 1, 1000]
        super().__init__(graph, critical_current_factors=Ic)

    def horizontal_junctions(self):
        return np.array([1, 0, -1, 0])

    def vertical_junctions(self):
        return  np.array([0, 1, 0, 1])


from __future__ import annotations

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection
from matplotlib.colors import Normalize, LogNorm
import matplotlib.animation as animation
import matplotlib.cm as cm
matplotlib.use("TkAgg")

from pyJJAsim.static_problem import StaticConfiguration
from pyJJAsim.time_evolution import TimeEvolutionResult

__all__ = ["Plot", "CircuitPlot", "CircuitMovie"]


class Plot():

    """Plots a circuit configuration.

    Allows one to show node quantities, junction quantities, face quantities and vortices.
        - Node quantities are shown as colors on the nodes
        - junction quantities are displayed by arrows whose length is proportional to the quantity value
        - face quantities are shown as colors on the faces
        - vortices are displayed by symbols (concentric rings, vorticity equals nr of rings, color shows sign).\

    Base class used by both CircuitPlot and CircuitMovie.\

    Node quantity options:
        - "phi": gauge depent phases
        - "U": node voltage or potential

    Junction quantity options:
        - "theta": gauge_invariant_phase_difference
        - "V": voltage
        - "I": current
        - "I_sup": supercurrent
        - "I_s": current sources
        - "EJ": josephson energy
        - "EM": magnetic_energy
        - "EC": capacitive energy
        - "Etot": total_energy

    Face quantity options:
        - "Phi: magnetic_flux
        - "n": vortex configuration
        - "J": face_current

    Attributes
    ----------
    config : StaticConfiguration
        some text!


    Methods
    -------
    test_method()
        Bla!

    """
    def __init__(self, config: StaticConfiguration | TimeEvolutionResult, time_point=0, show_vortices=True,
                 vortex_diameter=0.25, vortex_color=(0, 0, 0),
                 anti_vortex_color=(0.8, 0.1, 0.2), vortex_alpha=1, show_grid=True, grid_width=1,
                 grid_color=(0.3, 0.5, 0.9), grid_alpha=0.5, show_colorbar=True, show_arrows=True,
                 arrow_quantity="I", arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
                 arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1, arrow_color=(0.2, 0.4, 0.7),
                 arrow_alpha=1,  show_nodes=True, node_diameter=0.2,
                 node_face_color=(1,1,1), node_edge_color=(0, 0, 0), node_alpha=1, show_node_quantity=False,
                 node_quantity="phase", node_quantity_cmap=None, node_quantity_clim=(0, 1), node_quantity_alpha=1,
                 node_quantity_logarithmic_colors=False, show_face_quantity=False, face_quantity="n",
                 face_quantity_cmap=None, face_quantity_clim=(0, 1), face_quantity_alpha=1,
                 face_quantity_logarithmic_colors=False,
                 figsize=None, title=""):

        """
        Constructor for Plot and handling plot options.
        """
        self.config = config
        self.time_point = time_point

        if not isinstance(config, (StaticConfiguration, TimeEvolutionResult)):
            raise ValueError("config must be a StaticConfiguration or DynamicConfiguration object")

        self.show_vortices = show_vortices
        self.vortex_diameter = vortex_diameter
        self.vortex_color = vortex_color
        self.anti_vortex_color = anti_vortex_color
        self.vortex_alpha = vortex_alpha
        self.show_grid = show_grid
        self.grid_width = grid_width
        self.grid_color = grid_color
        self.grid_alpha = grid_alpha
        self.show_colorbar = show_colorbar
        self.show_arrows = show_arrows
        self.arrow_quantity = arrow_quantity

        self.arrow_width = arrow_width
        self.arrow_scale = arrow_scale
        self.arrow_headwidth = arrow_headwidth
        self.arrow_headlength = arrow_headlength
        self.arrow_headaxislength = arrow_headaxislength
        self.arrow_minshaft = arrow_minshaft
        self.arrow_minlength = arrow_minlength

        self.arrow_color = arrow_color
        self.arrow_alpha = arrow_alpha
        self.show_nodes = show_nodes
        self.node_diameter = node_diameter
        self.node_face_color = node_face_color
        self.node_edge_color = node_edge_color

        self.node_alpha = node_alpha
        self.show_node_quantity = show_node_quantity
        self.node_quantity = node_quantity
        self.node_quantity_cmap = node_quantity_cmap
        self.node_quantity_clim = node_quantity_clim
        self.node_quantity_alpha = node_quantity_alpha
        self.node_quantity_logarithmic_colors = node_quantity_logarithmic_colors

        self.show_face_quantity = show_face_quantity
        self.face_quantity = face_quantity
        self.face_quantity_cmap = face_quantity_cmap
        self.face_quantity_clim = face_quantity_clim
        self.face_quantity_alpha = face_quantity_alpha
        self.face_quantity_logarithmic_colors = face_quantity_logarithmic_colors

        self.figsize = figsize if figsize is not None else [6.4, 4.8]
        self.colorbar = None
        self.title = title

        self.fig = None
        self.ax = None

    _node_quantities = {
        "phi": 0, "phase": 0, "phases": 0, "U": 1, "potential": 1,
    }

    _junction_quantities = {
        "th": 0, "theta": 0, "phase_difference": 0,  "gauge_invariant_phase_difference": 0,
        "V": 1, "voltage": 1,
        "I": 3, "current": 3,
        "Isup": 4,  "I_sup": 4, "Isuper": 4, "I_super": 4, "supercurrent": 4, "super_current": 4,
        "I_s": 5, "Is": 5, "current_sources": 5,
        "EJ": 6, "josephson_energy": 6, "EM": 7, "magnetic_energy": 7, "EC": 8, "capacitive_energy": 8,
        "capacitance_energy": 8, "Etot": 9, "E_tot": 9, "ETot": 9, "total_energy": 9, "energy": 9,
    }

    _face_quantities = {
        "Phi": 0, "flux": 0, "magnetic_flux": 0,
        "n": 2, "vortices": 2, "vortex_configuration": 2, "face_current": 3, "J": 3,
    }

    def _get_lims(self):
        x, y = self.config.get_circuit().get_node_coordinates()
        xmin, xmax, ymin, ymax = np.min(x), np.max(x), np.min(y), np.max(y)
        dx, dy = xmax - xmin, ymax - ymin
        D = self.node_diameter * 0.5
        return xmin - 0.05 * dx - D, xmax + 0.05 * dx + D, ymin - 0.05 * dy - D, ymax + 0.05 * dy + D

    def _set_axes(self, ):
        xmin, xmax, ymin, ymax = self._get_lims()
        self.ax.set_xlim(xmin, xmax)
        self.ax.set_ylim(ymin, ymax)
        self.time_label = self.ax.annotate("", (xmin + 0.99 * (xmax - xmin), ymin + 0.98 * (ymax - ymin)), color=[1, 0.5, 0.2], ha='right', va='center')
        x0, y0, width, height = self.ax.get_position().bounds
        a_width = width * self.figsize[0]
        a_height = height * self.figsize[1]
        if a_width / (xmax - xmin) > a_height / (ymax - ymin):
            new_width = a_height * (xmax - xmin) / (ymax - ymin) / self.figsize[0]
            x0 = x0 + (width - new_width) / 2
            width = new_width
        else:
            new_height = a_width * (ymax - ymin) / (xmax - xmin) / self.figsize[1]
            y0 = y0 + (height - new_height) / 2
            height = new_height
        self.ax.set_position([x0, y0, width, height])

    def _is_static(self):
        return isinstance(self.config, StaticConfiguration)

    def _get_node_quantity(self):
        """
        Get node quantity (either "phi": gauge dependent phases or  "U": potential)
        """
        if isinstance(self.node_quantity, np.ndarray):
            return self.node_quantity.flatten()
        quantity = self._node_quantities[self.node_quantity]
        if quantity == 0:   # phi
            out = self.config.get_phi() if self._is_static() else self.config.get_phi(self.time_point)
            out = out.copy()
            out -= np.round(out / (np.pi * 2.0)).astype(out.dtype) * np.pi * 2.0
            return out
        if quantity == 1:   # U
            return self.config.get_U(self.time_point)

    def _get_junction_quantity(self):
        if isinstance(self.arrow_quantity, np.ndarray):
            return self.arrow_quantity.flatten()
        quantity = self._junction_quantities[self.arrow_quantity]
        if quantity == 0:   # theta
            out = self.config.get_theta() if self._is_static() else self.config.get_theta(self.time_point)
            out = out.copy()
            out -= np.round(out / (np.pi * 2.0)).astype(out.dtype) * np.pi * 2.0
            return out
        if quantity == 1:   # V
            return self.config.get_V(self.time_point)
        if quantity == 3:   # I
            return self.config.get_I() if self._is_static() else self.config.get_I(self.time_point)
        if quantity == 4:   # I
            return self.config.get_I() if self._is_static() else self.config.get_Isup(self.time_point)
        if quantity == 5:   # I_ext_J
            return self.config.problem._Is() if self._is_static() else self.config.problem._Is(self.time_point)
        if quantity == 6:   # EJ
            return self.config.get_EJ() if self._is_static() else self.config.get_EJ(self.time_point)
        if quantity == 7:   # EM
            return self.config.get_EM() if self._is_static() else self.config.get_EM(self.time_point)
        if quantity == 8:   # EC
            return self.config.get_EC(self.time_point)
        if quantity == 9:   # Etot
            return self.config.get_Etot() if self._is_static() else self.config.get_Etot(self.time_point)

    def _get_face_quantity(self):
        if isinstance(self.face_quantity, np.ndarray):
            return self.face_quantity.flatten()
        quantity = self._face_quantities[self.face_quantity]
        if quantity == 0:   # Phi
            return self.config.get_flux() if self._is_static() else self.config.get_flux(self.time_point)
        if quantity == 2:   # n
            return self.config.get_n() if self._is_static() else self.config.get_n(self.time_point)
        if quantity == 3:   # J
            return self.config.get_J() if self._is_static() else self.config.get_J(self.time_point)

    def _marker_scale_factor(self):
        xlim = self.ax.get_xlim()
        x0, y0, width, height = self.ax.get_position().bounds
        return (width * self.figsize[0]) / (xlim[1] - xlim[0]) *72


    def _plot_grid(self):
        x1, y1, x2, y2 = self.config.get_circuit().get_juncion_coordinates()
        self.ax.plot(np.stack((x1, x2)), np.stack((y1, y2)), color=self.grid_color,
                     alpha=self.grid_alpha, linewidth=self.grid_width, zorder=0)

    def _plot_nodes(self, node_quantity):
        x, y = self.config.get_circuit().get_node_coordinates()
        marker_size = self.node_diameter * self._marker_scale_factor()
        if not self.show_node_quantity:
            nodes_handle = self.ax.plot([x], [y], markeredgecolor=self.node_edge_color, markerfacecolor=self.node_face_color,
                                        markersize=marker_size, marker="o", alpha=self.node_alpha, zorder=2)
            nodes_handle = nodes_handle[0]
        else:
            cnorm = Normalize(*self.node_quantity_clim) if not self.node_quantity_logarithmic_colors else LogNorm(*self.node_quantity_clim)
            nodes_handle = self.ax.scatter(x.flatten(), y.flatten(), s=marker_size**2, c=node_quantity, cmap=self.node_quantity_cmap,
                                           edgecolors=self.node_edge_color, alpha=self.node_quantity_alpha, norm=cnorm)
            if self.show_colorbar:
                label = self.node_quantity if isinstance(self.node_quantity, str) else ""
                self.colorbar = plt.colorbar(cm.ScalarMappable(norm=cnorm, cmap=self.node_quantity_cmap), ax=self.ax, label=label)
        return nodes_handle

    def _plot_arrows(self, arrow_quantity):
        I = arrow_quantity * self.arrow_scale
        x1, y1, x2, y2 = self.config.get_circuit().get_juncion_coordinates()
        xq = x1 + 0.5 * (1 - I) * (x2 - x1)
        yq = y1 + 0.5 * (1 - I) * (y2 - y1)
        dxq, dyq = I * (x2 - x1), I * (y2 - y1)
        return self.ax.quiver(xq, yq, dxq, dyq, edgecolor=self.arrow_color, facecolor=self.arrow_color,
                   angles='xy', scale=1, scale_units='xy', width=self.arrow_width,
                   headwidth=self.arrow_headwidth, headlength=self.arrow_headlength,
                   headaxislength=self.arrow_headaxislength, minshaft=self.arrow_minshaft,
                   minlength=self.arrow_minlength, alpha=self.arrow_alpha, zorder=3)

    def _plot_faces(self, face_quantity):
        nodes = self.config.get_circuit().get_faces()
        x, y = self.config.get_circuit().get_node_coordinates()
        verts = [np.stack((x[n], y[n]), axis=-1) for n in nodes]
        cnorm = Normalize(*self.face_quantity_clim) if not self.face_quantity_logarithmic_colors else LogNorm(*self.face_quantity_clim)
        coll = PolyCollection(verts, array=face_quantity, edgecolors='none', cmap=self.face_quantity_cmap,
                              norm=cnorm, alpha=self.face_quantity_alpha, zorder=-1)
        faces_handle = self.ax.add_collection(coll)
        if self.show_colorbar and not (self.show_nodes and self.show_node_quantity):
            label = self.face_quantity if isinstance(self.face_quantity, str) else ""
            self.colorbar = plt.colorbar(cm.ScalarMappable(norm=cnorm, cmap=self.face_quantity_cmap), ax=self.ax, label=label)
        return faces_handle

    def _plot_vortices(self, n):
        marker_size = self.vortex_diameter * self._marker_scale_factor()
        xc, yc = self.config.get_circuit().get_face_centroids()
        ns = np.unique(n)
        vort_handles = []
        for ni in ns:
            if ni != 0:
                color = self.vortex_color if ni > 0 else self.anti_vortex_color
                na = np.abs(ni)
                for k in reversed(range(na)):
                    frac = (2 * k + 1) / (2 * na - 1)
                    p = self.ax.plot(xc[n == ni], yc[n == ni], markerfacecolor=color,
                                     markeredgecolor=color, marker="o", linestyle="",
                                     markersize=frac * marker_size, alpha=self.vortex_alpha, zorder=4)
                    vort_handles += p
                    if k > 0:
                        frac = (2 * k) / (2 * na - 1)
                        p = self.ax.plot(xc[n == ni], yc[n == ni], markerfacecolor=[1, 1, 1],
                                         markeredgecolor=[1, 1, 1], marker="o", linestyle="",
                                         markersize=frac * marker_size, alpha=self.vortex_alpha, zorder=4)
                        vort_handles += p
        return vort_handles

    def test_method(self):
        pass

class CircuitPlot(Plot):

    def __init__(self, config: StaticConfiguration | TimeEvolutionResult, time_point=0, show_vortices=True,
                 vortex_diameter=0.25, vortex_color=(0, 0, 0), anti_vortex_color=(0.8, 0.1, 0.2), vortex_alpha=1,
                 show_grid=True, grid_width=1, grid_color=(0.3, 0.5, 0.9), grid_alpha=0.5, show_colorbar=True,
                 show_arrows=True, arrow_quantity="I", arrow_width=0.005, arrow_scale=1, arrow_headwidth=3,
                 arrow_headlength=5, arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1,
                 arrow_color=(0.2, 0.4, 0.7), arrow_alpha=1, show_nodes=True, node_diameter=0.2,
                 node_face_color=(1, 1, 1), node_edge_color=(0, 0, 0), node_alpha=1, show_node_quantity=False,
                 node_quantity="phase", node_quantity_cmap=None, node_quantity_clim=(0, 1), node_quantity_alpha=1,
                 node_quantity_logarithmic_colors=False, show_face_quantity=False, face_quantity="n",
                 face_quantity_cmap=None, face_quantity_clim=(0, 1), face_quantity_alpha=1,
                 face_quantity_logarithmic_colors=False, figsize=None, title=""):

        super().__init__(config, time_point, show_vortices, vortex_diameter, vortex_color, anti_vortex_color,
                         vortex_alpha, show_grid, grid_width, grid_color, grid_alpha, show_colorbar, show_arrows,
                         arrow_quantity, arrow_width, arrow_scale, arrow_headwidth, arrow_headlength,
                         arrow_headaxislength, arrow_minshaft, arrow_minlength, arrow_color, arrow_alpha, show_nodes,
                         node_diameter, node_face_color, node_edge_color, node_alpha, show_node_quantity, node_quantity,
                         node_quantity_cmap, node_quantity_clim, node_quantity_alpha, node_quantity_logarithmic_colors,
                         show_face_quantity, face_quantity, face_quantity_cmap, face_quantity_clim, face_quantity_alpha,
                         face_quantity_logarithmic_colors, figsize, title)

        self.time_point = time_point

        if isinstance(config, TimeEvolutionResult):
            if not config.problem.store_time_steps[self.time_point]:
                raise ValueError("The requested timepoint from config to plot has not been stored during "
                                 "simulation (set with the config.store_time_steps property)")

    @staticmethod
    def _assert_single_configuration(data):
        if data.ndim >= 2:
            raise ValueError("must select single configuration")

    def make(self):
        self.fig, self.ax = plt.subplots(figsize=self.figsize)
        plt.title(self.title)
        self._set_axes()

        # get data
        n = self.config.get_n() if self._is_static() else self.config.get_n(self.time_point)[..., 0]
        self._assert_single_configuration(n)
        node_quantity, face_quantity, arrow_quantity = None, None, None
        if self.show_nodes and self.show_node_quantity:
            node_quantity = self._get_node_quantity()
            self._assert_single_configuration(node_quantity)
        if self.show_face_quantity:
            face_quantity = self._get_face_quantity()
            self._assert_single_configuration(face_quantity)
        if self.show_arrows:
            arrow_quantity = self._get_junction_quantity()
            self._assert_single_configuration(arrow_quantity)

        # plot data
        if self.show_grid:
            self._plot_grid()
        if self.show_face_quantity:
            self._plot_faces(face_quantity)
        if self.show_nodes:
            self._plot_nodes(node_quantity)
        if self.show_arrows:
            self._plot_arrows(arrow_quantity)
        if self.show_vortices:
            self._plot_vortices(n)

        # return handles
        if self.colorbar is not None:
            return self.fig, self.ax, self.colorbar
        else:
            return self.fig, self.ax


class CircuitMovie(Plot):

    def __init__(self, config: TimeEvolutionResult, problem_nr=0, time_points=None, show_vortices=True, vortex_diameter=0.25,
                 vortex_color=(0, 0, 0), anti_vortex_color=(0.8, 0.1, 0.2), vortex_alpha=1, show_grid=True,
                 grid_width=1, grid_color=(0.3, 0.5, 0.9), grid_alpha=0.5, show_colorbar=True, show_arrows=True,
                 arrow_quantity="I", arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
                 arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1, arrow_color=(0.2, 0.4, 0.7),
                 arrow_alpha=1, show_nodes=True, node_diameter=0.2, node_face_color=(1, 1, 1),
                 node_edge_color=(0, 0, 0), node_alpha=1, show_node_quantity=False, node_quantity="phase",
                 node_quantity_cmap=None, node_quantity_clim=(0, 1), node_quantity_alpha=1,
                 node_quantity_logarithmic_colors=False, show_face_quantity=False, face_quantity="n",
                 face_quantity_cmap=None, face_quantity_clim=(0, 1), face_quantity_alpha=1,
                 face_quantity_logarithmic_colors=False, figsize=None, animate_interval=5, title=""):

        super().__init__(config, time_points, show_vortices, vortex_diameter, vortex_color, anti_vortex_color, vortex_alpha,
                         show_grid, grid_width, grid_color, grid_alpha, show_colorbar, show_arrows, arrow_quantity,
                         arrow_width, arrow_scale, arrow_headwidth, arrow_headlength, arrow_headaxislength,
                         arrow_minshaft, arrow_minlength, arrow_color, arrow_alpha, show_nodes, node_diameter,
                         node_face_color, node_edge_color, node_alpha, show_node_quantity, node_quantity,
                         node_quantity_cmap, node_quantity_clim, node_quantity_alpha, node_quantity_logarithmic_colors,
                         show_face_quantity, face_quantity, face_quantity_cmap, face_quantity_clim, face_quantity_alpha,
                         face_quantity_logarithmic_colors, figsize, title)

        if self.time_point is None:
            self.time_point = np.ones(self.config.problem._Nt(), dtype=bool)
        if not (self.time_point.dtype in (bool, np.bool)):
            try:
                self.time_point = np.zeros(self.config.problem._Nt(), dtype=bool)
                self.time_point[time_points] = True
            except:
                raise ValueError("Invalid time_points; must be None, mask, slice or index array")
        self.problem_nr = problem_nr
        self.time_point &= self.config.problem.store_time_steps
        self.time_points_nz = np.flatnonzero(self.time_point)
        self.animate_interval = animate_interval
        self.faces_handle, self.nodes_handle, self.arrows_handle, self.vortices_handle = None, None, None, []
        self.item = None
        self.node_quantity_data = None
        self.face_quantity_data = None
        self.arrow_quantity_data = None
        self.n_data = None

    def show(self):
        self.fig, self.ax = plt.subplots(figsize=self.figsize)
        self._set_axes()
        plt.title(self.title)

        self.n_data = self.config.get_n()[:, self.problem_nr, self.config._time_point_index(self.time_point)]
        if self.show_nodes and self.show_node_quantity:
            self.node_quantity_data = self._get_node_quantity()[:, self.problem_nr, :]
        if self.show_face_quantity:
            self.face_quantity_data = self._get_face_quantity()[:, self.problem_nr, :]
        if self.show_arrows:
            self.arrow_quantity_data = self._get_junction_quantity()[:, self.problem_nr, :]
        if self.show_grid:
            self._plot_grid()
        time_point_list = np.arange(self.n_data.shape[-1], dtype=int)
        self.ani = animation.FuncAnimation(self.fig, self._animate, time_point_list,
                                           init_func=self._init, interval=self.animate_interval, blit=True)
        if self.colorbar is not None:
            return self.ani, self.fig, self.ax, self.colorbar
        else:
            return self.ani, self.fig, self.ax

    def _get_time_point_mask(self, time_point):
        mask = np.zeros(self.config.problem._Nt(), dtype=bool)
        mask[time_point] = True
        return mask

    def _animate(self, i):
        if self.show_face_quantity:
            self.faces_handle = self._update_faces(i)
        if self.show_nodes:
            self.nodes_handle = self._update_nodes(i)
        if self.show_arrows:
            self.arrows_handle = self._update_arrows(i)
        if self.show_vortices:
            self.vortex_handles =self._plot_vortices(self.n_data[:, i])
        self.time_label.set_text(str(self.time_points_nz[i]))
        handles = [self.nodes_handle, self.arrows_handle, self.faces_handle, self.time_label] + self.vortex_handles
        return [h for h in handles if h is not None]

    def _init(self):
        if self.show_face_quantity:
            self.faces_handle = self._plot_faces(self.face_quantity_data[:, 0])
        if self.show_nodes and self.show_node_quantity:
            self.nodes_handle = self._plot_nodes(self.node_quantity_data[:, 0])
        if self.show_arrows:
            self.arrows_handle = self._plot_arrows(self.arrow_quantity_data[:, 0])
        if self.show_vortices:
            self.vortex_handles = self._plot_vortices(self.n_data[:, 0])
        handles = [self.faces_handle, self.nodes_handle, self.arrows_handle, self.time_label] + self.vortex_handles
        return [h for h in handles if h is not None]

    def _update_faces(self, i):
        face_quantity = self.face_quantity_data[:, i]
        self.faces_handle.set_array(face_quantity)
        return self.faces_handle

    def _update_nodes(self, i):
        if self.show_node_quantity:
            if self.node_quantity_data.ndim >= 2:
                node_quantity = self.node_quantity_data[:, i]
            else:
                node_quantity = self.node_quantity_data
            self.nodes_handle.set_array(node_quantity)
        return self.nodes_handle

    def _update_arrows(self, i):
        I = self.arrow_quantity_data[:, i] * self.arrow_scale
        x1, y1, x2, y2 = self.config.get_circuit().get_juncion_coordinates()
        xq, yq = x1 + 0.5 * (1 - I) * (x2 - x1), y1 + 0.5 * (1 - I) * (y2 - y1)
        U, V = I * (x2 - x1), I * (y2 - y1)
        self.arrows_handle.set_UVC(U, V)
        self.arrows_handle.X = xq
        self.arrows_handle.Y = yq
        self.arrows_handle.XY = np.concatenate((xq[:, None], yq[:, None]), axis=1)
        self.arrows_handle._offsets = self.arrows_handle.XY
        return self.arrows_handle




from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "pyJJAsim/README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '2.1.3'
DESCRIPTION = 'Circuit simulator for josephson junctions and passive components'
LONG_DESCRIPTION = 'A package that allows simulating transport properties of arrays of josephson junctions.' \
                   'Features include time evolution and finding static configurations, external magnetic fields and' \
                   'thermal fluctuations.'

# Setting up
setup(
    name="pyjjasim",
    version=VERSION,
    author="Martijn Lankhorst",
    author_email="<m.lankhorst89@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=['numpy', 'scipy', 'matplotlib'],
    keywords=['python', 'josephson_junction_array', 'circuit', 'simulation'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
    ]
)


from __future__ import annotations

import numpy as np
import scipy
import scipy.sparse.linalg
import scipy.optimize

from pyJJAsim.josephson_circuit import Circuit
from pyJJAsim.static_problem import DefaultCPR
from pyJJAsim.static_problem import StaticConfiguration, StaticProblem

__all__ = ["TimeEvolutionProblem", "TimeEvolutionResult", "AnnealingProblem"]


class TimeEvolutionProblem:
    """
    Define multiple time evolution problems with varying parameters in a Josephson Junction Circuit circuit.

     - All problems are computed in one call (generally much faster than computing individual problems).
     - Use self.compute() to obtain a TimeEvolutionResult object containing the resulting time evolution.

    Physical parameters:       symbol  default
     - time_step               dt      0.05
     - time_step_count         Nt      1000
     - current_phase_relation  cp      DefaultCPR()

    Problem space parameters: symbol  default   shape
     - frustration            f       0.0       array broadcast compatible to (Nf, W, Nt)
                                                or f(i) -> broadcast compatible to (Nf, W) for i in range(Nt)
     - current_sources        Is      0.0       array broadcast compatible to (Nj, W, Nt)
                                                or Is(i) -> broadcast compatible to (Nj, W) for i in range(Nt)
     - voltage_sources        Vs      0.0       array broadcast compatible to (Nj, W, Nt)
                                                or Vs(i) -> broadcast compatible to (Nj, W) for i in range(Nt)
     - temperature            T       0.0       array broadcast compatible to (Nj, W, Nt)
                                                or T(i) -> broadcast compatible to (Nj, W) for i in range(Nt)

    Where W is problem_count.

    Store parameters:     default    type                           Needed for:
     - store_time_steps   None       None (represents all points)
                                     or array in range(Nt)
                                     or mask of shape (Nt,)
     - store_theta        True       bool                           phi, EJ, ...
     - store_voltage      True       bool
     - store_current      True       bool                           Phi, ...

    The store parameters allow one to control what quantities are stored at what timesteps. Only the base quantities
    theta, voltage and current can be stored in the resulting TimeEvolutionResult; other quantities like energy
    or magnetic_flux are computed based on these. A list is given which quantities require what base quantities
    to be stored (see documentation of TimeEvolutionResult for more information).

    Initial condition parameters:   default  type
     - config_at_minus_1            None     StaticConfiguration    initial condition at timestep=-1
                                             or None                 -> set to self.get_static_problem(t=0,n=z).compute()
     - config_at_minus_2            None     StaticConfiguration    initial condition at timestep=-2
                                             or None                 -> set to self.get_static_problem(t=0,n=z).compute()

    Methods
     - compute() -> TimeEvolutionResult
     - get_static_problem(time_step=0) -> StaticProblem
     - get_problem_count()
     - get_circuit()
     - get_time_step()
     - get_time_step_count()
     - get_current_phase_relation()
     - get_phase_zone()
     - get_frustration()
     - get_current_sources()
     - get_voltage_sources()
     - get_temperature()
     - get_store_time_steps()
     - get_store_theta()
     - get_store_voltage()
     - get_store_current()


    """

    def __init__(self, circuit: Circuit, time_step=0.05, time_step_count=1000,
                 current_phase_relation=DefaultCPR(),
                 frustration=0.0, current_sources=0.0,
                 voltage_sources=0.0, temperature=0.0,
                 store_time_steps=None, store_theta=True, store_voltage=True, store_current=True,
                 config_at_minus_1: np.ndarray = None,
                 config_at_minus_2: np.ndarray = None):

        self.circuit = circuit
        self.time_step = time_step
        self.time_step_count = time_step_count
        self.current_phase_relation = current_phase_relation

        def get_prob_cnt(x):
            s = np.array(x(0) if hasattr(x, "__call__") else x).shape
            return s[1] if len(s) > 1 else 1

        self.problem_count = max(get_prob_cnt(frustration),
                                 get_prob_cnt(current_sources), get_prob_cnt(voltage_sources),
                                 get_prob_cnt(temperature))
        Nj, Nf, W, Nt = self.circuit._Nj(), self.circuit._Nf(), self.get_problem_count(), self.get_time_step_count()
        self.frustration = frustration if hasattr(frustration, "__call__") else \
            self._broadcast(np.array(frustration), (Nf, W, Nt))
        self.current_sources = current_sources if hasattr(current_sources, "__call__") else \
            self._broadcast(np.array(current_sources), (Nj, W, Nt))
        self.voltage_sources = voltage_sources if hasattr(voltage_sources, "__call__") else \
            self._broadcast(np.array(voltage_sources), (Nj, W, Nt))
        self.temperature = temperature if hasattr(temperature, "__call__") else \
            self._broadcast(np.array(temperature), (Nj, W, Nt))

        self.store_time_steps = np.ones(self._Nt(), dtype=bool)
        self.store_time_steps = self._to_time_point_mask(store_time_steps)
        self.store_theta = store_theta
        self.store_voltage = store_voltage
        self.store_current = store_current

        self.config_at_minus_1 = np.zeros((Nj, W), dtype=np.double) if config_at_minus_1 is None else config_at_minus_1
        self.config_at_minus_2 = np.zeros((Nj, W), dtype=np.double) if config_at_minus_2 is None else config_at_minus_2
        self.prepared_theta_s = np.zeros((Nj, W), dtype=np.double)

    def get_static_problem(self, vortex_configuration, problem_nr=0, time_step=0) -> StaticProblem:
        return StaticProblem(self.circuit, current_sources=self._Is(time_step)[:, problem_nr].copy(),
                             frustration=self._f(time_step)[:, problem_nr].copy(),
                             vortex_configuration=vortex_configuration,
                             current_phase_relation=self.current_phase_relation)

    def get_problem_count(self):
        return self.problem_count

    def get_circuit(self) -> Circuit:
        return self.circuit

    def get_time_step(self):
        return self.time_step

    def get_time_step_count(self):
        return self.time_step_count

    def get_current_phase_relation(self):
        return self.current_phase_relation

    def get_phase_zone(self):
        return 0

    def get_frustration(self):
        return self.frustration

    def get_current_sources(self):
        return self.current_sources

    def get_net_sourced_current(self, time_step):
        M = self.get_circuit().get_cut_matrix()
        return 0.5 * np.sum(np.abs((M @ self._Is(time_step))), axis=0)

    def get_voltage_sources(self):
        return self.voltage_sources

    def get_temperature(self):
        return self.temperature

    def get_store_time_steps(self):
        return self.store_time_steps

    def get_store_theta(self):
        return self.store_theta

    def get_store_voltage(self):
        return self.store_voltage

    def get_store_current(self):
        return self.store_current

    def get_time(self):
        return np.arange(self._Nt(), dtype=np.double) * self._dt()

    def get_time_at_stored(self):
        return self.get_time()[self.store_time_steps]

    def compute(self) -> TimeEvolutionResult:
        """
        Compute time evolution on an Josephson Circuit.

        Requires an initial configuration; step0_init. Must be StaticConfiguration or None.
        If None; it is set to self.to_static_problem().compute()

        If the circuit has capacitance, requires a second initial configuration; step1_init. If
        this is set to None, it will be assigned the value of step0_init.

        If there is no inductance and algorithm=0; the initial condition must obey A(th-g) = 0.
        If this is not obeyed, it is automatically projected to obey the constraint. Due to numerical
        rounding, it will slowly drift away from this condition. Use rounding_flux_drift_correction=True
        to apply projection every 100 timesteps.

        """
        if self.get_circuit()._has_mixed_inductance():
            return time_evolution_algo_1(self)
        else:
            return time_evolution_algo_0(self)

    def __str__(self):
        return "time evolution problem: " + \
               "\n\ttime: " + self.time_step_count.__str__() + " steps of " + self.time_step.__str__() + \
               "\n\tcurrent sources: " + self.current_sources.__str__() + \
               "\n\tvoltage sources: " + self.voltage_sources.__str__() + \
               "\n\tfrustration: " + self.frustration.__str__() + \
               "\n\ttemperature: " + self.temperature.__str__() + \
               "\n\tcurrent-phase relation: " + self.current_phase_relation.__str__()

    def _Nt(self):
        return self.time_step_count

    def _Nt_s(self):
        return np.asscalar(np.sum(self.store_time_steps))

    def _dt(self):
        return self.time_step

    def _f(self, time_step) -> np.ndarray:  # (Nf, W), read-only
        return np.broadcast_to(self.frustration(time_step), (self.circuit._Nf(), self.get_problem_count())) \
            if hasattr(self.frustration, "__call__") else self.frustration[:, :, time_step]

    def _Is(self, time_step) -> np.ndarray: # (Nj, W), read-only
        return np.broadcast_to(self.current_sources(time_step), (self.circuit._Nj(), self.get_problem_count()))\
            if hasattr(self.current_sources, "__call__") else self.current_sources[:, :, time_step]

    def _Vs(self, time_step) -> np.ndarray: # (Nj, W), read-only
        return np.broadcast_to(self.voltage_sources(time_step), (self.circuit._Nj(), self.get_problem_count()))\
            if hasattr(self.voltage_sources, "__call__") else self.voltage_sources[:, :, time_step]

    def _T(self, time_step) -> np.ndarray:  # (Nj, W), read-only
        return np.broadcast_to(self.temperature(time_step), (self.circuit._Nj(), self.get_problem_count()))\
            if hasattr(self.temperature, "__call__") else self.temperature[:, :, time_step]

    def _theta_s(self, time_step) -> np.ndarray:    # (Nj, W), read-only
        # warning: must be called exactly once at every timestep in order, starting below zero.
        if time_step < 0:
            self.prepared_theta_s = self._Vs(0) * (time_step * self._dt())
        else:
            self.prepared_theta_s = self.prepared_theta_s + self._Vs(time_step) * self._dt()
            return self.prepared_theta_s

    def _cp(self, theta) -> np.ndarray:   # (Nj, W)
        # theta -> (Nj, W)
        Ic = self.get_circuit()._Ic().diag(force_as_vector=True, vector_length=self.get_circuit()._Nj())[:, None]
        return self.current_phase_relation.eval(Ic, theta)

    def _dcp(self, theta) -> np.ndarray:  # (Nj, W)
        Ic = self.get_circuit()._Ic().diag(force_as_vector=True, vector_length=self.get_circuit()._Nj())[:, None]
        return self.current_phase_relation.d_eval(Ic, theta)

    def _icp(self, theta) -> np.ndarray:  # (Nj, W)
        Ic = self.get_circuit()._Ic().diag(force_as_vector=True, vector_length=self.get_circuit()._Nj())[:, None]
        return self.current_phase_relation.i_eval(Ic, theta)

    def _broadcast(self, x, shape):
        x_shape = np.array(x).shape
        x = x.reshape(x_shape + (1,) * (len(shape) - len(x_shape)))
        return np.broadcast_to(x, shape)

    def _to_time_point_mask(self, time_points):
        # time_points:  None (represents all points)
        #               or slice
        #               or array in range(Nt)
        #               or mask of shape (Nt,)
        if time_points is None:
            time_points = self.store_time_steps
        time_points = np.array(time_points)
        if not (time_points.dtype in (bool, np.bool)):
            try:
                x = np.zeros(self._Nt(), dtype=bool)
                x[time_points] = True
                time_points = x
            except:
                raise ValueError("Invalid store_time_steps; must be None, mask, slice or index array")
        return time_points

def time_evolution_algo_0(problem: TimeEvolutionProblem) -> TimeEvolutionResult:
    out = TimeEvolutionResult(problem)

    circuit = problem.get_circuit()
    Nj, W = circuit._Nj(), problem.get_problem_count()
    dt = problem._dt()
    if circuit._has_mixed_inductance():
        raise ValueError("Time evolution algorithm 0 does not work with mixed inductance "
                         "(some loops have no inductance while others do. Use algorithm 1.")
    store_th, store_I, store_V = problem.store_theta, problem.store_current, problem.store_voltage

    A = circuit.get_cycle_matrix()
    AT = A.T
    Rv = 1 / (dt * circuit._R().diag(force_as_vector=True, vector_length=Nj)[:, None])
    Cv = circuit._C().diag(force_as_vector=True, vector_length=Nj)[:, None] / (dt ** 2)
    Cprev, C0, Cnext = Cv, -2.0 * Cv - Rv, Cv + Rv

    if circuit._has_inductance():
        L = problem.circuit._L().matrix(Nj)
        L_sw_fact = scipy.sparse.linalg.factorized(A @ L @ AT)
    Asq_fact = scipy.sparse.linalg.factorized(A @ scipy.sparse.diags(1.0 / Cnext[:, 0], 0) @ AT)

    theta_next = problem.config_at_minus_1
    theta = problem.config_at_minus_2

    for i in range(problem._Nt()):
        Is, T, theta_s, f = problem._Is(i), problem._T(i), problem._theta_s(i), problem._f(i)

        rand = np.random.randn(Nj, W) if i % 3 == 0 else rand[np.random.permutation(Nj), :]
        fluctuations = ((2.0 * T * Rv) ** 0.5) * rand

        theta_prev = theta.copy()
        theta = theta_next.copy()
        if circuit._has_inductance():
            y = AT @ L_sw_fact(A @ (theta + theta_s + L @ Is) + 2 * np.pi * f)
            theta_next = -(problem._cp(theta) + fluctuations - Is + C0 * theta + Cprev * theta_prev + y) / Cnext
            I = Is - y if store_I else None
        else:
            x = (problem._cp(theta) + fluctuations - Is + C0 * theta + Cprev * theta_prev) / Cnext
            theta_next = -x + (AT @ Asq_fact(A @ (x - theta_s) - 2 * np.pi * f)) / Cnext
            I = (x + theta_next) * Cnext + Is if store_I else None
        V = (theta_next - theta) / dt if store_V else None
        if problem.store_time_steps[i]:
            out._update([theta_next if problem.store_theta else None, V, I])
    return out

def time_evolution_algo_1(problem: TimeEvolutionProblem) -> TimeEvolutionResult:
    """

    """

    out = TimeEvolutionResult(problem)

    circuit = problem.circuit
    Nj, Nf, W = circuit._Nj(),  circuit._Nf(), problem.get_problem_count()
    dt = problem._dt()

    store_th, store_I, store_V = problem.store_theta, problem.store_current, problem.store_voltage

    A = circuit.get_cycle_matrix()
    M = circuit._Mr().A
    Rv = 1 / (dt * circuit._R().diag(force_as_vector=True, vector_length=Nj)[:, None])
    Cv = circuit._C().diag(force_as_vector=True, vector_length=Nj)[:, None] / (dt ** 2)
    Cprev, C0, Cnext = Cv, -2.0 * Cv - Rv, Cv + Rv

    L = problem.circuit._L().matrix(Nj)
    L_mask = circuit._get_mixed_inductance_mask()
    A1 = A[~L_mask, :]
    A2 = A[L_mask, :]

    Cnext_mat = scipy.sparse.diags(Cnext[:, 0], 0)
    matrix = scipy.sparse.vstack([M @ Cnext_mat, A1 @ L @ Cnext_mat, A2]).tocsc()
    m_fact = scipy.sparse.linalg.factorized(matrix)

    theta = problem.config_at_minus_2
    theta_next = problem.config_at_minus_1

    for i in range(problem._Nt()):
        Is, T, theta_s, f = problem._Is(i), problem._T(i), problem._theta_s(i), problem._f(i)

        # optimization to only generate new gaussian noise every three timesteps.
        # On the other timesteps, the last generated noise is shuffled.
        rand = np.random.randn(Nj, W) if i % 3 == 0 else rand[np.random.permutation(Nj), :]
        fluctuations = ((2.0 * T * Rv) ** 0.5) * rand

        theta_prev = theta.copy()
        theta = theta_next.copy()
        y = problem._cp(theta) + fluctuations + C0 * theta + Cprev * theta_prev
        F = np.concatenate([M @ (y - Is), A1 @ (theta + theta_s + L @ y) + 2 * np.pi * f[~L_mask],
                            A2 @ theta_s + 2 * np.pi * f[L_mask]], axis=0)
        theta_next = -m_fact(F)

        x = Cprev * theta_prev + C0 * theta + Cnext * theta_next + problem._cp(theta) + fluctuations
        matrix = scipy.sparse.vstack([M, A @ L]).tocsc()
        print(scipy.linalg.norm(matrix @ x + np.concatenate([-M @ Is, A @ (theta + theta_s) + 2 * np.pi * f], axis=0)))

        if problem.store_time_steps[i]:
            out._update([theta_next if problem.store_theta else None,
                         (theta_next - theta) / dt if store_V else None,
                         y + Cnext * theta_next if store_I else None])
    return out


class TimeEvolutionResult:
    """
    Represents data of simulated time evolution(s) on a Josephson circuit.

    It is defined by a problem, circuit and any of the quantities theta, current and voltage.
    These must be of shape (*problem.get_shape(), Nj, Nt_stored). Here Nj is the junction count
    and Nt_stored is the number of time steps that are stored during simulation.

    One can query several properties of the circuit configurations:

     (property)                           (symbol)           (needs)    (shape)
     - phases                             phi                th         (pr_shape, Nn, Nt_stored)
     - gauge_invariant_phase_difference   theta                         (pr_shape, Nj, Nt_stored)
     - vortex_configuration               n                  th         (pr_shape, Nj, Nt_stored)
     - junction_current                   I                             (pr_shape, Nj, Nt_stored)
     - supercurrent                       Isup               th         (pr_shape, Nj, Nt_stored)
     - path_current                       J                  I          (pr_shape, Np, Nt_stored)
     - magnetic_flux (through faces)      flux               I          (pr_shape, Nn, Nt_stored)
     - junction_voltage                   V                             (pr_shape, Nj, Nt_stored)
     - node_voltage                       U                  V          (pr_shape, Nn, Nt_stored)
     - josephson_energy                   EJ                 th         (pr_shape, Nj, Nt_stored)
     - magnetic_energy                    EM                 I          (pr_shape, Nj, Nt_stored)
     - capacitor_energy                   EC                 V          (pr_shape, Nj, Nt_stored)
     - total_energy                       Etot               th,(I),(V) (pr_shape, Nj, Nt_stored)

    A property query is done with te command .get_[symbol](select_time_points=None)

    where:
        select_time_points      None (represents all stored points)
                                or array in range(Nt)
                                or mask of shape (Nt,)

    TimeEvolutionResult only store theta, current and voltage data at specified timepoints,
    other queried properties must be calculated from that. The above table shows which
    quantities need to be stored to be able to query a certain property. (parenthesis
    mean quantities may be needed).

    Thermal quantities are based only on theta, and represent "thermal averages", in the
    sense that derivatives are computed on theta over the data at the queried timepoints
    (nót the stored timepoints). This naturally smooths the quantity, which filters the
    thermal noise present at nonzero temperature.

    """

    def __init__(self, problem: TimeEvolutionProblem):
        self.problem = problem
        Nj, W, Nt_s = problem.circuit._Nj(), self.get_problem_count(), problem._Nt_s()
        self.theta = np.zeros((Nj, W, Nt_s), dtype=np.double) if problem.store_theta else None
        self.voltage = np.zeros((Nj, W, Nt_s), dtype=np.double) if problem.store_theta else None
        self.current = np.zeros((Nj, W, Nt_s), dtype=np.double) if problem.store_theta else None
        self.store_point = 0
        s = self.problem.store_time_steps.astype(int)
        self.time_point_indices = np.cumsum(s) - s

    def _update(self, data):
        th, V, I = data[0], data[1], data[2]
        if th is not None:
            self.theta[:, :, self.store_point] = th
        if V is not None:
            self.voltage[:, :, self.store_point] = V
        if I is not None:
            self.current[:, :, self.store_point] = I
        if (th is not None) or (th is not None) or (th is not None):
            self.store_point += 1

    def _th(self, time_point) -> np.ndarray:
        if self.theta is None:
            raise ValueError("Cannot query theta; quantity is not stored during time evolution simulation.")
        return self.theta[:, :, self._time_point_index(time_point)]

    def _V(self, time_point) -> np.ndarray:
        if self.theta is None:
            raise ValueError("Cannot query voltage; quantity is not stored during time evolution simulation.")
        return self.voltage[:, :, self._time_point_index(time_point)]

    def _I(self, time_point) -> np.ndarray:
        if self.theta is None:
            raise ValueError("Cannot query current; quantity is not stored during time evolution simulation.")
        return self.current[:, :, self._time_point_index(time_point)]

    def _time_point_index(self, time_points):
        if time_points is None:
            time_points = self.problem.store_time_steps
        if not np.all(self.problem.store_time_steps[time_points]):
            raise ValueError("Queried a timepoint that is not stored during time evolution simulation.")
        return self.time_point_indices[time_points]

    def get_problem_count(self):
        return self.problem.get_problem_count()

    def get_circuit(self) -> Circuit:
        return self.problem.get_circuit()

    def select_static_configuration(self, prob_nr, time_step) -> StaticConfiguration:
        if self.theta is None:
            raise ValueError("Theta not stored; cannot select static configuration.")
        problem = StaticProblem(self.get_circuit(), current_sources=self.problem._Is(time_step)[:, prob_nr],
                                frustration=self.problem._f(time_step)[:, prob_nr],
                                vortex_configuration=self.get_n(time_step)[:, prob_nr],
                                current_phase_relation=self.problem.current_phase_relation)
        return StaticConfiguration(problem, self.theta[:, prob_nr, time_step])

    def get_phi(self, select_time_points=None) -> np.ndarray:
        M = self.get_circuit()._Mr().A
        Mrsq = M @ M.T
        Z = np.zeros((1, self.get_problem_count()), dtype=np.double)
        func = lambda tp: np.concatenate((scipy.sparse.linalg.spsolve(Mrsq, M @ self._th(tp)), Z), axis=0)
        return self._select(select_time_points, self.get_circuit()._Nn(), func)

    def get_theta(self, select_time_points=None) -> np.ndarray:
        return self._select(select_time_points, self.get_circuit()._Nj(), self._th)

    def get_n(self, select_time_points=None) -> np.ndarray:
        A = self.get_circuit().get_cycle_matrix()
        func = lambda tp:  -A @ np.round(self._th(tp) / (2.0 * np.pi))
        return self._select(select_time_points, self.get_circuit()._Nf(), func).astype(int)

    def get_EJ(self, select_time_points=None) -> np.ndarray:
        func = lambda tp: self.problem._icp(self._th(tp))
        return self._select(select_time_points, self.get_circuit()._Nj(), func)

    def get_I(self, select_time_points=None) -> np.ndarray:
        return self._select(select_time_points, self.get_circuit()._Nj(), self._I)

    def get_Isup(self, select_time_points=None) -> np.ndarray:
        func = lambda tp: self.problem._cp(self._th(tp))
        return self._select(select_time_points, self.get_circuit()._Nj(), func)

    def get_J(self, select_time_points=None) -> np.ndarray:
        A = self.get_circuit().get_cycle_matrix()
        func = lambda tp: scipy.sparse.linalg.spsolve(A @ A.T,  A @ self._I(tp))
        return self._select(select_time_points, self.get_circuit()._Nf(), func)

    def get_flux(self, select_time_points=None) -> np.ndarray:
        Nj, Nf = self.get_circuit()._Nj(), self.get_circuit()._Nf()
        A = self.get_circuit().get_cycle_matrix()
        func = lambda tp: A @ (self.get_circuit()._L().matrix(Nj) @ self._I(tp))
        return self._select(select_time_points, Nf, func)

    def get_EM(self, select_time_points=None) -> np.ndarray:
        Nj = self.get_circuit()._Nj()
        func = lambda tp: 0.5 * self.get_circuit()._L().matrix(Nj) @ (self._I(tp) ** 2)
        return self._select(select_time_points, Nj, func)

    def get_V(self, select_time_points=None):
        return self._select(select_time_points, self.get_circuit()._Nj(), self._V)

    def get_U(self, select_time_points=None):
        M = self.get_circuit()._Mr().A
        Mrsq = M @ M.T
        Z = np.zeros((1, self.get_problem_count()), dtype=np.double)
        func = lambda tp: np.concatenate((scipy.sparse.linalg.spsolve(Mrsq, M @ self._V(tp)), Z), axis=0)
        return self._select(select_time_points, self.get_circuit()._Nn(), func)

    def get_EC(self, select_time_points=None):
        Nj = self.get_circuit()._Nj()
        C = self.get_circuit()._C().diag(force_as_vector=True, vector_length=Nj)
        func = lambda tp: 0.5 * C[:, None] * self._V(tp) ** 2
        return self._select(select_time_points, Nj, func)

    def get_Etot(self, select_time_points=None) -> np.ndarray:
        return self.get_EJ(select_time_points) + self.get_EM(select_time_points) + \
               self.get_EC(select_time_points)

    def plot(self, time_point=0, show_vortices=True, vortex_diameter=0.25, vortex_color=(0, 0, 0),
             anti_vortex_color=(0.8, 0.1, 0.2), vortex_alpha=1, show_grid=True, grid_width=1,
             grid_color=(0.3, 0.5, 0.9), grid_alpha=0.5, show_colorbar=True, show_arrows=True,
             arrow_quantity="I", arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
             arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1, arrow_color=(0.2, 0.4, 0.7),
             arrow_alpha=1, show_nodes=True, node_diameter=0.2,
             node_face_color=(1, 1, 1), node_edge_color=(0, 0, 0), node_alpha=1, show_node_quantity=False,
             node_quantity="phase", node_quantity_cmap=None, node_quantity_clim=(0, 1), node_quantity_alpha=1,
             node_quantity_logarithmic_colors=False, show_face_quantity=False, face_quantity="n",
             face_quantity_cmap=None, face_quantity_clim=(0, 1), face_quantity_alpha=1,
             face_quantity_logarithmic_colors=False, figsize=None, title="", **kwargs):

        from circuit_visualize import CircuitPlot

        return CircuitPlot(self, time_point=time_point, show_vortices=show_vortices, vortex_diameter=vortex_diameter,
                         vortex_color=vortex_color, anti_vortex_color=anti_vortex_color,
                         vortex_alpha=vortex_alpha, show_grid=show_grid, grid_width=grid_width,
                         grid_color=grid_color, grid_alpha=grid_alpha, show_colorbar=show_colorbar,
                         show_arrows=show_arrows,
                         arrow_quantity=arrow_quantity, arrow_width=arrow_width, arrow_scale=arrow_scale,
                         arrow_headwidth=arrow_headwidth, arrow_headlength=arrow_headlength,
                         arrow_headaxislength=arrow_headaxislength, arrow_minshaft=arrow_minshaft,
                         arrow_minlength=arrow_minlength, arrow_color=arrow_color,
                         arrow_alpha=arrow_alpha, show_nodes=show_nodes, node_diameter=node_diameter,
                         node_face_color=node_face_color, node_edge_color=node_edge_color,
                         node_alpha=node_alpha, show_node_quantity=show_node_quantity,
                         node_quantity=node_quantity, node_quantity_cmap=node_quantity_cmap,
                         node_quantity_clim=node_quantity_clim, node_quantity_alpha=node_quantity_alpha,
                         node_quantity_logarithmic_colors=node_quantity_logarithmic_colors,
                         show_face_quantity=show_face_quantity, face_quantity=face_quantity,
                         face_quantity_cmap=face_quantity_cmap, face_quantity_clim=face_quantity_clim,
                         face_quantity_alpha=face_quantity_alpha,
                         face_quantity_logarithmic_colors=face_quantity_logarithmic_colors,
                         figsize=figsize, title=title, **kwargs).make()

    def animate(self, problem_nr=0, time_points=None, show_vortices=True,
                vortex_diameter=0.25, vortex_color=(0, 0, 0), anti_vortex_color=(0.8, 0.1, 0.2),
                vortex_alpha=1, show_grid=True, grid_width=1,
                grid_color=(0.3, 0.5, 0.9), grid_alpha=0.5, show_colorbar=True, show_arrows=True, arrow_quantity="I",
                arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
                arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1, arrow_color=(0.2, 0.4, 0.7),
                arrow_alpha=1, show_nodes=True, node_diameter=0.2,
                node_face_color=(1, 1, 1), node_edge_color=(0, 0, 0), node_alpha=1, show_node_quantity=False,
                node_quantity="phase", node_quantity_cmap=None, node_quantity_clim=(0, 1), node_quantity_alpha=1,
                node_quantity_logarithmic_colors=False,
                show_face_quantity=False, face_quantity="n", face_quantity_cmap=None,
                face_quantity_clim=(0, 1), face_quantity_alpha=1,
                face_quantity_logarithmic_colors=False, figsize=None,
                animate_interval=5, title=""):

        from pyJJAsim.circuit_visualize import CircuitMovie

        return CircuitMovie(self, problem_nr=problem_nr, time_points=time_points, show_vortices=show_vortices,
                          vortex_diameter=vortex_diameter, vortex_color=vortex_color,
                          anti_vortex_color=anti_vortex_color, vortex_alpha=vortex_alpha,
                          show_grid=show_grid, grid_width=grid_width,
                          grid_color=grid_color, grid_alpha=grid_alpha,
                          show_colorbar=show_colorbar, show_arrows=show_arrows,
                          arrow_quantity=arrow_quantity, arrow_width=arrow_width, arrow_scale=arrow_scale,
                          arrow_headwidth=arrow_headwidth, arrow_headlength=arrow_headlength,
                          arrow_headaxislength=arrow_headaxislength, arrow_minshaft=arrow_minshaft,
                          arrow_minlength=arrow_minlength, arrow_color=arrow_color,
                          arrow_alpha=arrow_alpha, show_nodes=show_nodes, node_diameter=node_diameter,
                          node_face_color=node_face_color, node_edge_color=node_edge_color,
                          node_alpha=node_alpha, show_node_quantity=show_node_quantity,
                          node_quantity=node_quantity, node_quantity_cmap=node_quantity_cmap,
                          node_quantity_clim=node_quantity_clim, node_quantity_alpha=node_quantity_alpha,
                          node_quantity_logarithmic_colors=node_quantity_logarithmic_colors,
                          show_face_quantity=show_face_quantity, face_quantity=face_quantity,
                          face_quantity_cmap=face_quantity_cmap, face_quantity_clim=face_quantity_clim,
                          face_quantity_alpha=face_quantity_alpha,
                          face_quantity_logarithmic_colors=face_quantity_logarithmic_colors,
                          figsize=figsize, animate_interval=animate_interval, title=title).show()

    def __str__(self):
        return "time evolution configuration: (" + ("th" + self.theta.shape.__str__() + ", ") * (
                    self.theta is not None) + \
               ("I" + self.current.shape.__str__() + ", ") * (self.current is not None) + \
               ("V" + self.voltage.shape.__str__()) * (self.current is not None) + ")" + \
               "\nproblem: " + self.problem.__str__() + \
               "\ncircuit: " + self.get_circuit().__str__()

    def _select(self, select_time_points, N, func):
        select_time_points = np.flatnonzero(self.problem._to_time_point_mask(select_time_points))
        W = self.get_problem_count()
        out = np.zeros((N, W, len(select_time_points)), dtype=np.double)
        for i, tp in enumerate(select_time_points):
            out[:, :, i] = func(tp)
        return out


class AnnealingProblem:

    """
    Anneals a circuit by gradually lowering the temperature, with the goal for finding a stationairy
    state with reasonably low energy. The temperature profile is computed automatically based on the
    measured vortex mobility during the run.

     - Does interval_count iterations of interval_steps timeseteps.
     - The first iteration is done at T=start_T
     - During each iteration it measures the average vortex mobility. If the target mobility is exceeded,
       the temperature is divided by T_factor, otherwise it is multiplied by it.
     - The target vortex mobility v_t at iter i  is v_t(i) = v * ((N - i)/N) ** 1.5, so goes from v to 0.
     - At the end it does some steps at T=0 to ensure it is settled in a stationairy state.
     - vortex mobility is defined as abs(n(i+1) - n(i))) / dt averaged over space and time.
    """
    def __init__(self, circuit: Circuit, time_step=0.5, interval_steps=10,
                 interval_count=1000, vortex_mobility=0.001,
                 frustration=0.0, current_sources=0, problem_count=1,
                 start_T=1.0, T_factor=1.03):
        self.circuit = circuit
        self.time_step = time_step
        self.interval_steps = interval_steps
        self.interval_count = interval_count
        self.vortex_mobility = vortex_mobility
        self.current_sources = current_sources
        self.frustration = frustration
        self.problem_count = problem_count
        self.T =start_T * np.ones((1, self.problem_count, 1))
        self.T_factor = T_factor

    def get_vortex_mobility(self, n):
        Nf = self.circuit.face_count()
        return np.sum(np.sum(np.abs(np.diff(n, axis=2)), axis=2), axis=0) / (Nf * self.time_step * (self.interval_count - 1))

    def temperature_adjustment(self, vortex_mobility, iteration):
        v = self.vortex_mobility
        upper = v[iteration] if (np.array(v)).size == self.interval_count else \
            v * ((self.interval_count - iteration) / self.interval_count) ** 1.5
        factor = (vortex_mobility > upper) * (1/self.T_factor) +  (vortex_mobility <= upper) * self.T_factor
        self.T *= factor[..., None]

    def compute(self):
        """
        Execute the annealing procedure. Returns:
        vortex_configurations   (Nf, problem_count)
        energies                (problem_count,)
        status
        configurations          (problem_count,) list of StaticConfiguration objects
        temperature_profiles    (interval_count, problem_count)
        """

        f = np.array(self.frustration)[:, None, None]
        th = np.zeros((self.circuit.junction_count(), self.problem_count))
        prob = TimeEvolutionProblem(self.circuit, time_step_count=self.interval_steps, time_step=self.time_step,
                                    frustration=f, current_sources=self.current_sources, temperature=self.T,
                                    store_current=False, store_voltage=False)
        T = np.zeros((self.interval_count, self.problem_count))
        for i in range(self.interval_count):
            prob.temperature = self.T * np.ones((1, 1, self.interval_steps))
            prob.config_at_minus_1 = th
            out = prob.compute()
            vortex_configurations = out.get_n()
            vortex_mobility = self.get_vortex_mobility(vortex_configurations)
            self.temperature_adjustment(vortex_mobility, i)
            th = out.get_theta()[..., -1]
            # stat_conf = out.select_static_configuration(prob_nr=0, time_step=self.interval_steps-1)
            # print(stat_conf.get_error())
            T[i, :] = self.T[0, :, 0]
        prob.temperature = np.zeros((1, 1, self.interval_steps))
        prob.time_step /=2
        for i in range(1 + (self.interval_count//5)):
            prob.config_at_minus_1 = th
            out = prob.compute()
        vortex_configurations = out.get_n()[:, :, -1]
        data = [prob.get_static_problem(vortex_configurations[:, p], problem_nr=0, time_step=0).compute()
                for p in range(self.problem_count)]
        configurations = [d[0] for d in data]
        energies = np.array([np.mean(d[0].get_Etot()) for d in data])
        status = np.array([d[1] for d in data])
        return vortex_configurations, energies, status, configurations, T



















import time
import numpy as np
import scipy
import scipy.sparse.linalg
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("TkAgg")

from josephson_circuit import HoneycombArray, SquareArray


array = SquareArray(50, 50)
Nn, Nj, Nf = array._Nnr(), array._Nj(), array._Nf()
array.set_inductance_factors(1)

M = array._Mr().A
A = array.get_cycle_matrix()

m = scipy.sparse.vstack([M, A ]).tocsc()

W = 1000
x = np.random.rand(Nn, W)
y = np.random.rand(Nj, W)

r = scipy.sparse.diags(np.random.rand(Nj), 0)

Mf = scipy.sparse.linalg.factorized(M @ r @ M.T)
mf = scipy.sparse.linalg.factorized(m @ m.T)

tic = time.perf_counter()
M.T @ Mf(M @ y)
print(time.perf_counter() - tic)

tic = time.perf_counter()
mf(y)
print(time.perf_counter() - tic)


tic = time.perf_counter()
fluctuations = np.random.randn(Nj, W)
fluctuations = np.random.randn(Nj, W)
fluctuations = np.random.randn(Nj, W)
print(time.perf_counter() - tic)

tic = time.perf_counter()
fluctuations = np.random.randn(Nj, W)
perm = np.random.permutation(Nj)
fluctuations = fluctuations[perm, :]
perm = np.random.permutation(Nj)
fluctuations = fluctuations[perm, :]
print(time.perf_counter() - tic)